// Needs fixed prulib 
// http://e2e.ti.com/support/arm/sitara_arm/f/791/t/239735.aspx
// https://github.com/beagleboard/am335x_pru_package/issues/3

// This module is the routines needed to emulate a MFM disk. The emulated
// disk data is the raw clock and data MFM bit stream for each track.
// The ARM handles reading and writing a cylinder of track data to DDR 
// memory buffer. PRU 1 converts the MFM bits to Pulse Width Modulation (PWM)
// words PRU 0 uses to generate the MFM bitstream. PRU 1 also converts the
// MFM delta transitions times from PRU 0 to the MFM bitstream and stores
// it back in DDR memory. PRU 0 handles outputting the PWM words to the PWM
// generator, reading the delta words from the hardware and handling the
// various drive control lines. If the host seeks PRU 0 interrupts the ARM
// to let it fetch the next cylinder.
//

// Copyright 2014 David Gesswein.
// This file is part of MFM disk utilities.
//
// 11/09/14 DJG Added write buffer and delays to handle slow flash memory
// 09/06/14 DJG Fixed deadlock between shutting down PRU 0 and seeking
//    to next cylinder
//
// MFM disk utilities is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// MFM disk utilities is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with MFM disk utilities.  If not, see <http://www.gnu.org/licenses/>.
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <inttypes.h>
#include <signal.h>
#include <pthread.h>
#include <semaphore.h>

// MFM program include files
#include <prussdrv.h>
#include <pruss_intc_mapping.h>

#include "msg.h"
#include "pru_setup.h"
#include "emu_tran_file.h"
#include "parse_cmdline.h"

#include "cmd.h"

#define ARRAYSIZE(x) (sizeof(x) / sizeof(x[0]))

// Pick best timer clock
#ifdef CLOCK_MONOTONIC_RAW
#define CLOCK CLOCK_MONOTONIC_RAW
#else
#define CLOCK CLOCK_MONOTONIC
#endif

// File syste read/write threads
pthread_t read_thread;
pthread_t write_thread;

// This is a circular buffer holding tracks to write to the file. The flash
// can't keep up with random writes so this is needed to prevent timeouts when
// the writes block. Buffer is empty when get==put
struct {
   int drive;        // Drive, cylinder and head data is for, drive -1 signals
   int cyl;          // write thread to exit
   int head;
   int size;         // Size of track data
   char *buf;        // Track data
} *track_buffer;
// Circular buffer indexes
volatile int track_buffer_get;
volatile int track_buffer_put;

// Semaphore to let write thread know more data is available
sem_t write_sem;

// Log for run information
FILE *log_file;
// time() this program was started at
time_t start_time;

// TODO Find a new home for this
//
// This routine will either set the CPU speed to maximum or restore it
// back to the previous setting. The normal governor didn't boost the
// processor sufficiently to prevent getting behind in processing.
//
// restore: 0 to set speed to maximum, 1 to restore value from previous
//    call
int set_restore_max_cpu_speed(int restore) {
   FILE *file;
   static char governor[100];
   static int freq_changed = 0;
   char maxfreq[100];

   if (!restore) {
      file = fopen("/sys/devices/system/cpu/cpu0/cpufreq/scaling_governor", "r");
      if (file == NULL) {
         return -1;
      }
      if (fscanf(file, "%100s", governor) < 1) {
         return -1;
      }
      file = fopen("/sys/devices/system/cpu/cpu0/cpufreq/scaling_governor", "w");
      if (file == NULL) {
         return -1;
      }
      if (fprintf(file,"userspace") < 1) {
         return -1;
      }
      fclose(file);
      file = fopen("/sys/devices/system/cpu/cpu0/cpufreq/cpuinfo_max_freq", "r");
      if (file == NULL) {
         return -1;
      }
      if (fscanf(file, "%100s", maxfreq) < 1) {
         return -1;
      }
      fclose(file);

      file = fopen("/sys/devices/system/cpu/cpu0/cpufreq/scaling_setspeed", "w");
      if (file == NULL) {
         return -1;
      }
      if (fprintf(file,maxfreq) < 1) {
         return -1;
      }
      fclose(file);
      freq_changed = 1;
   } else {
      if (freq_changed) {
         file = fopen("/sys/devices/system/cpu/cpu0/cpufreq/scaling_governor", "w");
         if (file == NULL) {
            return -1;
         }
         if (fprintf(file,governor) < 1) {
            return -1;
         }
         fclose(file);
      }
      freq_changed = 0;
   }
   return 0;
}


// This routine is for cleaning up when shutting down. Its installed as an
// atexit routine and called by SIGINT handler.
void shutdown(void)
{
   static int called = 0;
   time_t stop_time;
   struct timespec tv_start;
   double shutdown_start, shutdown_stop;

   if (called)
      return;
   called = 1;

   // Time how long it takes from shutdown request to file flushed so users
   // can see how its doing versus the capacitor run time. 
   clock_gettime(CLOCK, &tv_start);
   shutdown_start = tv_start.tv_sec + tv_start.tv_nsec / 1e9;

   set_restore_max_cpu_speed(1);

   // Command PRU code to termiate
   pru_write_word(MEM_PRU0_DATA,PRU0_EXIT, 1);

   // Wait for read and write threads to complete. File flushed by
   // write thread
   pthread_join(write_thread, NULL);
   pthread_join(read_thread, NULL);

   // Wait for PRU to signal its done
   pru_shutdown();

   clock_gettime(CLOCK, &tv_start);
   shutdown_stop = tv_start.tv_sec + tv_start.tv_nsec / 1e9;
   fprintf(log_file,"   Shutdown time %.1f seconds\n",
      shutdown_stop - shutdown_start);

   stop_time = time(NULL);
   fprintf(log_file,"Emulation stopped %.24s, duration %.3f days\n",ctime(&stop_time),
      (stop_time - start_time)/60.0/60.0/24.0);
}

// SIGINT/ control-c handler. Call our shutdown and exit
void shutdown_signal(void)
{
   shutdown();
   exit(1);
}

// Get number of entries used in track buffer
// size: Total number of entries in buffer
// return: Number of entries used. Last entry can't
//         be used to maximum return is size-1
int track_buffers_used(int size) {
   int used = track_buffer_put - track_buffer_get;

   if (used < 0) {
      used += size;
   }

   return used;
}

// This routine update the buffer parameters and put pointer and posts
// write thread semaphore to write data in buffer
//
// buffer_count: Number of buffers
// drive: drive number data is for
// cyl: cylinder to write to
// head: head/track to write to
// size: size of track data
void update_buffer(int buffer_count, int drive, int cyl, int head, int size) {
   int next_put;

   next_put = (track_buffer_put + 1) % buffer_count;
   if (next_put == track_buffer_get) {
      msg(MSG_INFO, "Track buffer full\n");
   }
   // If no free buffers wait until one is free
   while (next_put == track_buffer_get) {
      usleep(10000);
   }
   track_buffer[track_buffer_put].drive = drive;
   track_buffer[track_buffer_put].cyl = cyl;
   track_buffer[track_buffer_put].head = head;
   track_buffer[track_buffer_put].size = size;
   track_buffer_put = next_put;

   sem_post(&write_sem);
}

// Get the cylinder data and send to PRU. The cylinder is read from the file
// and if the track buffer has later data it is copied to the cylinder data.
// The data is then sent to the PRU
//
// drive_params: Drive parameters
// drive: Drive number
// cyl: Cylinder
// cyl_size: Cylinder size in bytes including headers
// track_size: Track size in bytes including header
// data: Cylinder data
void send_PRU_cyl_data(DRIVE_PARAMS *drive_params, int drive, int cyl,
      int cyl_size, int track_size, uint8_t *data) {
   int get_hold, index;

   // We need to have get index from before we read from file to ensure we
   // check all data that hasn't been written before the read. Put can't
   // change during this routine.
   get_hold = track_buffer_get;
   emu_file_read_cyl(drive_params->fd[drive],
         &drive_params->emu_file_info[drive], cyl, data,  cyl_size);
   // We need to go from get pointer to put pointer to ensure the last
   // data we move is the latest. We may move data we overwrite again.
   index = get_hold;
   while (index != track_buffer_put) {
      // If our drive and cylinder than move the track data to the correct
      // location in cylinder buffer
      if (track_buffer[index].drive == drive && track_buffer[index].cyl == cyl) {
printf("Updating from buffer %d get_hold %d put %d\n", index, get_hold, track_buffer_put);
         memcpy(&data[track_size * track_buffer[index].head],
               track_buffer[index].buf, track_buffer[index].size);
      }
      index = (index + 1) % drive_params->buffer_count;
   }

   pru_write_mem(MEM_DDR, data, cyl_size, drive*DDR_DRIVE_BUFFER_MAX_SIZE);
}

// This thread writes a cylinder of data to the shared DDR memory for the
// PRU's. It waits for an interrupt, writes the buffer back to disk if dirty,
// then reads the next cylinder of data. The data to write is sent through
// track buffers to another thread to actually write to prevent host computer
// timeouts when the write blocks. We delay between writes a linear
// increasing delay as the buffers get full. The flash memory can't keep up
// with random writes.
// TODO: Speed up memory copy.
//
// arg: drive_params pointer
static void *emu_proc(void *arg)
{
   DRIVE_PARAMS *drive_params = arg;
   // Buffer sizes for cylinder and track for each emulated drive
   int cyl_size[MAX_DRIVES];
   int track_size[MAX_DRIVES];
   // Current and new cylinder for each drive. Initial values force transfer
   // of cylinder 0 at startup
   int cyl[MAX_DRIVES] = {-1,-1};
   int new_cyl[MAX_DRIVES] = {0,0};
   // Buffer for the drive data
   uint8_t *data[MAX_DRIVES];
   // PRU register to decode for debug
   uint32_t b;
   // How long we took to respond to PRU request for next cylinder
   double seek_time, max_seek_time = 0;
   // How long we should delay
   float delay_time;
   // Bit flag for which tracks need to be written
   int dirty;
   int i;
   int done = 0;
   int trk;
   // Minimum free buffers
   int min_free_buf;
   // Buffer variables
   int num_used_buf, num_free_buf, init_used_buf;
   uint32_t total_seeks = 0, total_writes = 0;

   min_free_buf = drive_params->buffer_count;
   // Setup our variables
   for (i = 0; i < drive_params->num_drives; i++) {
      track_size[i] = drive_params->emu_file_info[i].track_data_size_bytes +
         drive_params->emu_file_info[i].track_header_size_bytes;
      cyl_size[i] = track_size[i] * drive_params->emu_file_info[i].num_head;
      data[i] = malloc(cyl_size[i]);
      if (data[i] == NULL) {
         msg(MSG_FATAL, "Cylinder buffer malloc failed, size %d\n",
            cyl_size[i]);
         exit(1);
      }
   }

   while (!done) {
      for (i = 0; i < drive_params->num_drives; i++) {
         if (new_cyl[i] != cyl[i]) {
            total_seeks++;
            cyl[i] = new_cyl[i];
            // Get new cylinder data and send to PRU
            send_PRU_cyl_data(drive_params, i, cyl[i], cyl_size[i],
                  track_size[i], data[i]);
         }
      }
      pru_exec_cmd(CMD_START, 0);
      seek_time = pru_read_word(MEM_PRU0_DATA,PRU0_SEEK_TIME) / 200e6 * 1e3;
      if (seek_time > max_seek_time) {
         max_seek_time = seek_time;
      }
      msg(MSG_INFO,"  Waiting, seek time %.1f ms max %.1f min free buffers %d\n",
         seek_time, max_seek_time, min_free_buf);
      // Wait for request from PRU
      prussdrv_pru_wait_event (PRU_EVTOUT_0);
      prussdrv_pru_clear_event (PRU_EVTOUT_0, PRU0_ARM_INTERRUPT);
      // If no data to write we don't delay
      delay_time = 0;
      num_used_buf = 0;
      num_free_buf = -1;
      init_used_buf = track_buffers_used(drive_params->buffer_count);

      for (i = 0; i < drive_params->num_drives; i++) {
         dirty = pru_read_word(MEM_PRU1_DATA, PRU1_DRIVE0_TRK_DIRTY +
            i*PRU_WORD_SIZE_BYTES);
         new_cyl[i] = pru_read_word(MEM_PRU0_DATA,PRU0_DRIVE0_CUR_CYL +
            i*PRU_WORD_SIZE_BYTES);
         if (new_cyl[i] == -1) {
            done = 1;
         }
         b = pru_read_word(MEM_PRU0_DATA,PRU0_CUR_SELECT_HEAD);
         if (cyl[i] != new_cyl[i]) {
            msg(MSG_INFO,"  Drive %d Cyl %d->%d select %d, head %d dirty %x\n",
               i, cyl[i], new_cyl[i],
               (((b >> 22) & 0x3) | ((b >> 24) & 0xc)) ^ 0xf, 
               ((b >> 2) & 0xff) ^ 0xf, dirty);
         }
         if (dirty) {
            total_writes++;
            pru_write_word(MEM_PRU1_DATA,PRU1_DRIVE0_TRK_DIRTY +
               i*PRU_WORD_SIZE_BYTES, 0);
            for (trk = 0; trk < drive_params->emu_file_info[i].num_head; trk++) {
               // For each dirty track read the data from PRU, put in
               // buffer and tell writer it has more data.
               if (dirty & (1 << trk)) {
                  pru_read_mem(MEM_DDR, track_buffer[track_buffer_put].buf,
                     track_size[i], track_size[i]*trk +
                     i*DDR_DRIVE_BUFFER_MAX_SIZE);

                  num_used_buf = track_buffers_used(drive_params->buffer_count);
                  // -1 is because last buffer can't be used
                  num_free_buf = drive_params->buffer_count - num_used_buf - 1;

                  if (num_free_buf < min_free_buf) {
                     min_free_buf = num_free_buf;
                  }
                  update_buffer(drive_params->buffer_count, i, cyl[i], trk, track_size[i]);

                  // Do linear delay based on number of buffers full
                  // We will do one delay after all data transfered
                  delay_time += num_used_buf * drive_params->buffer_time;
               }
            }
         }
      }
      if (num_free_buf != -1) {
         msg(MSG_INFO, "Free buffers %d,%d delay %.3f\n", num_free_buf, 
            init_used_buf, delay_time);
      }
      // If delay small or only couple buffers full before last transfer then
      // don't delay.
      if (delay_time > 2e-3 && init_used_buf > 2) {
         // If sum of delays greater than maximum only delay maximum
         // Assume maximum is when controller timeouts will happen
         if (delay_time > drive_params->buffer_max_time) {
            delay_time = drive_params->buffer_max_time;
         }

         msg(MSG_INFO, "Actual delay %.3f\n", delay_time);
         usleep((int) (delay_time * 1e6));
      }
   }
   // Tell PRU we saw the exiting flag
   pru_write_word(MEM_PRU0_DATA,PRU0_DRIVE0_CUR_CYL, 0);
   update_buffer(drive_params->buffer_count, -1, 0, 0, 0);

   fprintf(log_file,"   Max seek time %.1f ms min free buffers %d, %u seeks %u writes\n",
      max_seek_time, min_free_buf, total_seeks, total_writes);

   return NULL;
}

// This thread writes data from the buffers to disk
//
// arg: drive_params pointer
static void *emu_proc_write(void *arg)
{
   DRIVE_PARAMS *drive_params = arg;
   int i;

   while (1) {
      // wait until data available
      sem_wait(&write_sem);

      // drive -1 signals we are done
      if (track_buffer[track_buffer_get].drive == -1) {
         break;
      }

      emu_file_rewrite_track(
         drive_params->fd[track_buffer[track_buffer_get].drive],
         &drive_params->emu_file_info[track_buffer[track_buffer_get].drive],
         track_buffer[track_buffer_get].cyl,
         track_buffer[track_buffer_get].head,
         track_buffer[track_buffer_get].buf,
         track_buffer[track_buffer_get].size);
      track_buffer_get = (track_buffer_get + 1) % drive_params->buffer_count;
   }

   // Done with files
   for (i = 0; i < MAX_DRIVES; i++) {
      emu_file_close(drive_params->fd[i], 0);
   }

   return NULL;
}

// Main routine. 
int main(int argc, char *argv[])
{
   // Size of shared memory in bytes
   int ddr_mem_size;
   int i;
   DRIVE_PARAMS drive_params;
   uint32_t zeros[EMU_FILE_TRACK_DATA_BYTES/4];
   int cyl,head;
   int select_map[] = 
      {0, GPIO_SELECT1, GPIO_SELECT2, GPIO_SELECT3, GPIO_SELECT4};
   int max_buffer = 0;
   struct sched_param params;

   memset(zeros, 0, sizeof(zeros));
   // Find out what we should do
   parse_cmdline(argc, argv, &drive_params);

   if (drive_params.initialize) {
      if (drive_params.num_drives != 1) {
         msg(MSG_FATAL, "Only one filename may be specified when initializing\n");
         exit(1);
      }
      drive_params.cmdline = parse_print_cmdline(&drive_params, 0);
      drive_params.fd[0] = emu_file_write_header(drive_params.filename[0],
         drive_params.num_cyl, drive_params.num_head, drive_params.cmdline,
         drive_params.note);
      for (cyl = 0; cyl < drive_params.num_cyl; cyl++) {
         for (head = 0; head < drive_params.num_head; head++) {
            emu_file_write_track_bits(drive_params.fd[0], zeros,
               EMU_FILE_TRACK_DATA_BYTES/4, cyl, head);
         }
      }
      emu_file_close(drive_params.fd[0], 1);
   }
   // Initialize PRU
   ddr_mem_size = pru_setup(2);
   if (ddr_mem_size == -1) {
      exit(1);
   }

   for (i = 0; i < MAX_DRIVES; i++) {
      // 0 for drive always selected, 1-4 for select 1-4
      if (i < drive_params.num_drives) {
         if (drive_params.drive[i] < ARRAYSIZE(select_map)) {
            pru_write_word(MEM_PRU0_DATA,PRU0_DRIVE0_SELECT +
               i*PRU_WORD_SIZE_BYTES,
               select_map[drive_params.drive[i]]);
         } else {
            msg(MSG_FATAL, "Illegal drive select %d\n", drive_params.drive);
            exit(1);
         }
      } else {
         pru_write_word(MEM_PRU0_DATA,PRU0_DRIVE0_SELECT + i*
            (PRU0_DRIVE1_SELECT - PRU0_DRIVE0_SELECT), 0);
      }
   }
   // And start our code
   if (prussdrv_exec_program(0, "prucode0.bin") != 0) {
      msg(MSG_FATAL, "Unable to execute prucode0.bin\n");
      exit(1);
   }

   if (prussdrv_exec_program(1, "prucode1.bin") != 0) {
      msg(MSG_FATAL, "Unable to execute prucode1.bin\n");
      exit(1);
   }

   log_file = fopen("logfile.txt","a");
   start_time = time(NULL);
   fprintf(log_file,"Emulation started %.24s\n",ctime(&start_time));
   fflush(log_file);

   // this messes up IEP running/accessing registers
   // Cleanup when we exit
   signal(SIGINT,(__sighandler_t) shutdown_signal);
   atexit(shutdown);
   // Heads must be zero for unused drives
   for (i = 0; i < MAX_DRIVES; i++) {
      pru_write_word(MEM_PRU0_DATA,PRU0_DRIVE0_NUM_HEAD + i*PRU_WORD_SIZE_BYTES,
         0);
   }

   // Send parameters for drives emulating to PRU
   for (i = 0; i < drive_params.num_drives; i++) {
      drive_params.fd[i] = emu_file_read_header(drive_params.filename[i],
         &drive_params.emu_file_info[i], 1);
      if (drive_params.emu_file_info[i].note != NULL &&
         strlen(drive_params.emu_file_info[i].note) != 0) {
         msg(MSG_INFO, "Drive %d note: %s\n", i, 
            drive_params.emu_file_info[i].note);
      }

      msg(MSG_INFO,"Drive %d num cyl %d num head %d track len %d\n", i,
         drive_params.emu_file_info[i].num_cyl, 
         drive_params.emu_file_info[i].num_head,
         drive_params.emu_file_info[i].track_data_size_bytes);

      pru_write_word(MEM_PRU0_DATA,PRU0_DRIVE0_NUM_HEAD + i*PRU_WORD_SIZE_BYTES,
         drive_params.emu_file_info[i].num_head);
      pru_write_word(MEM_PRU0_DATA,PRU0_DRIVE0_NUM_CYL + i*PRU_WORD_SIZE_BYTES,
         drive_params.emu_file_info[i].num_cyl);
      pru_write_word(MEM_PRU1_DATA,PRU1_DRIVE0_TRACK_HEADER_BYTES +
         i*PRU_WORD_SIZE_BYTES, drive_params.emu_file_info[i].track_header_size_bytes);
      pru_write_word(MEM_PRU1_DATA,PRU1_DRIVE0_TRACK_DATA_BYTES +
         i*PRU_WORD_SIZE_BYTES, drive_params.emu_file_info[i].track_data_size_bytes);
      if (drive_params.emu_file_info[i].track_data_size_bytes > max_buffer) {
         max_buffer = drive_params.emu_file_info[i].track_data_size_bytes;
      }
   }
   track_buffer = malloc(drive_params.buffer_count * sizeof(*track_buffer));
   if (track_buffer == NULL) {
      msg(MSG_FATAL, "Track buffer malloc failed\n");
      exit(1);
   }
   memset(track_buffer, 0, drive_params.buffer_count * sizeof(*track_buffer));
   for (i = 0; i < drive_params.buffer_count; i++) {
      track_buffer[i].buf = malloc(max_buffer);
      if (track_buffer[i].buf == NULL) {
         msg(MSG_FATAL, "Track buffer buf malloc failed\n");
         exit(1);
      }
   }
   track_buffer_put = 0;
   track_buffer_get = 0;

   sem_init(&write_sem, 0, 0);

   if (pthread_create(&read_thread, NULL, &emu_proc, &drive_params)
      != 0) {
      msg(MSG_FATAL, "Unable to create read thread\n");
      exit(1);
   }

   if (pthread_create(&write_thread, NULL, &emu_proc_write, &drive_params)
      != 0) {
      msg(MSG_FATAL, "Unable to create read thread\n");
      exit(1);
   }

   params.sched_priority = sched_get_priority_max(SCHED_FIFO);
   if (pthread_setschedparam(read_thread, SCHED_FIFO, &params) != 0) {
      msg(MSG_ERR, "Unable to set thread priority\n");
   }
#if 1
   // TODO: Should this drop speed when idle?
   // This is needed to keep up with the data. Without it we will take
   // more than 2 revolutions per track due to the default governor not
   // increasing the CPU speed enough. We switch frequently between busy and
   // sleeping.
   if (set_restore_max_cpu_speed(0)) {
      msg(MSG_ERR, "Unable to set CPU to maximum speed\n");
   }
#endif
   sleep(1);
   // This is debugging stuff to print state of data from PRU
   struct {
      int loc;
      MEM_TYPE mem_type;
      char *txt;
      uint32_t last;
   } locs[] = {
      {PRU1_BAD_PATTERN_COUNT, MEM_PRU1_DATA, "bad pattern count %d\n", 123},
      {PRU0_RQUEUE_UNDERRUN, MEM_PRU0_DATA, "Read queue underrun %d\n", 123},
      {PRU0_WQUEUE_OVERRUN, MEM_PRU0_DATA, "Write queue overrun %d\n", 123},
      {PRU0_ECAP_OVERRUN, MEM_PRU0_DATA, "Ecapture overrun %d\n", 123},
      {PRU0_HEAD_SELECT_GLITCH_COUNT, MEM_PRU0_DATA, "glitch count %d\n", 123},
      {PRU0_HEAD_SELECT_GLITCH_VALUE, MEM_PRU0_DATA, "glitch value %x\n", 123},
      {PRU_TEST0, MEM_PRU0_DATA, "0:test 0 %x\n", 123},
      {PRU_TEST1, MEM_PRU0_DATA, "0:test 1 %x\n", 123},
      {PRU_TEST2, MEM_PRU0_DATA, "0:test 2 %x\n", 123},
      {PRU_TEST3, MEM_PRU0_DATA, "0:test 3 %x\n", 123},
      {PRU_TEST4, MEM_PRU0_DATA, "0:test 4 %x\n", 123},
      {PRU_TEST0, MEM_PRU1_DATA, "1:test 0 %x\n", 123},
      {PRU_TEST1, MEM_PRU1_DATA, "1:test 1 %x\n", 123},
      {PRU_TEST2, MEM_PRU1_DATA, "1:test 2 %x\n", 123},
      {PRU_TEST3, MEM_PRU1_DATA, "1:test 3 %x\n", 123},
      {PRU_TEST4, MEM_PRU1_DATA, "1:test 4 %x\n", 123}
   };
   while (1) {
      uint32_t a;
#if 1
      for (i = 0; i < ARRAYSIZE(locs); i++) {
         a = pru_read_word(locs[i].mem_type,locs[i].loc);
         if (a != locs[i].last) {
            printf(locs[i].txt,a);
            locs[i].last = a;
         }
      } 
#endif
#if 1
      static uint32_t b, last_b = 9;
      b = pru_read_word(MEM_PRU0_DATA,PRU0_CUR_SELECT_HEAD);
      if (b != last_b) {
         printf("select %d head %d\n",
            (((b >> 22) & 0x3) | ((b >> 24) & 0xc)) ^ 0xf,
            ((b >> 2) & 0xff) ^ 0xf);
         last_b = b;
      } 
#endif
      //printf("PC %x %x\n", pru_get_pc(0), pru_get_pc(1));
      usleep(1000000);
   }

   // Not reached, exit called to terminate
   return (0);
}

